@@ -3,22 +3,20 @@ from __future__ import annotations
 import dataclasses
 from collections.abc import Callable
 from enum import StrEnum
-from typing import Any, Optional, TypeVar, cast
+from typing import Any, TypeVar, cast
 
-_ClassT = TypeVar("_ClassT")
+from .annotation import SerialBase, SerialInspector
+
+ClassT = TypeVar("ClassT")
 
 
-@dataclasses.dataclass
-class _SerialClassData:
+@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
+class _SerialClassData(SerialBase):
     unconverted_keys: set[str] = dataclasses.field(default_factory=set)
     unconverted_values: set[str] = dataclasses.field(default_factory=set)
     to_string_values: set[str] = dataclasses.field(default_factory=set)
     parse_require: set[str] = dataclasses.field(default_factory=set)
-    named_type_path: Optional[str] = None
-    # Tracks if this data was provided as a decorator to the type.
-    # This is used to track "proper types" which are appropriate
-    # for serialization and/or dynamic discovery
-    from_decorator: bool = False
+    named_type_path: str | None = None
 
 
 EMPTY_SERIAL_CLASS_DATA = _SerialClassData()
@@ -26,12 +24,13 @@ EMPTY_SERIAL_CLASS_DATA = _SerialClassData()
 
 def serial_class(
     *,
-    unconverted_keys: Optional[set[str]] = None,
-    unconverted_values: Optional[set[str]] = None,
-    to_string_values: Optional[set[str]] = None,
-    parse_require: Optional[set[str]] = None,
-    named_type_path: Optional[str] = None,
-) -> Callable[[_ClassT], _ClassT]:
+    unconverted_keys: set[str] | None = None,
+    unconverted_values: set[str] | None = None,
+    to_string_values: set[str] | None = None,
+    parse_require: set[str] | None = None,
+    named_type_path: str | None = None,
+    is_dynamic_allowed: bool = False,
+) -> Callable[[ClassT], ClassT]:
     """
     An additional decorator to a dataclass that specifies serialization options.
 
@@ -55,7 +54,7 @@ def serial_class(
         The type_spec type-path to this type. This applies only to named types.
     """
 
-    def decorate(orig_class: _ClassT) -> _ClassT:
+    def decorate(orig_class: ClassT) -> ClassT:
         cast(Any, orig_class).__unc_serial_data = _SerialClassData(
             unconverted_keys=unconverted_keys or set(),
             unconverted_values=unconverted_values or set(),
@@ -63,17 +62,20 @@ def serial_class(
             parse_require=parse_require or set(),
             named_type_path=named_type_path,
             from_decorator=True,
+            is_dynamic_allowed=is_dynamic_allowed,
         )
         return orig_class
 
     return decorate
 
 
-class SerialClassDataInspector:
+class SerialClassDataInspector(SerialInspector[ClassT]):
     def __init__(
         self,
+        parsed_type: type[ClassT],
         current: _SerialClassData,
     ) -> None:
+        super().__init__(parsed_type, current)
         self.current = current
 
     def has_unconverted_key(self, key: str) -> bool:
@@ -88,16 +90,8 @@ class SerialClassDataInspector:
     def has_parse_require(self, key: str) -> bool:
         return key in self.current.parse_require
 
-    @property
-    def from_decorator(self) -> bool:
-        return self.current.from_decorator
 
-    @property
-    def named_type_path(self) -> Optional[str]:
-        return self.current.named_type_path
-
-
-def _get_merged_serial_class_data(type_class: type[Any]) -> _SerialClassData | None:
+def get_merged_serial_class_data(type_class: type[Any]) -> _SerialClassData | None:
     base_class_data = (
         cast(_SerialClassData, type_class.__unc_serial_data)
         if hasattr(type_class, "__unc_serial_data")
@@ -106,28 +100,30 @@ def _get_merged_serial_class_data(type_class: type[Any]) -> _SerialClassData | N
     if base_class_data is None:
         return None
 
+    # IMPROVE: We should cache this result on the type
     if type_class.__bases__ is not None:
         for base in type_class.__bases__:
-            curr_base_class_data = _get_merged_serial_class_data(base)
+            curr_base_class_data = get_merged_serial_class_data(base)
             if curr_base_class_data is not None:
-                if base_class_data is None:
-                    base_class_data = _SerialClassData()
-                base_class_data.unconverted_keys |= (
-                    curr_base_class_data.unconverted_keys
+                base_class_data = dataclasses.replace(
+                    base_class_data,
+                    unconverted_keys=base_class_data.unconverted_keys
+                    | curr_base_class_data.unconverted_keys,
+                    unconverted_values=base_class_data.unconverted_values
+                    | curr_base_class_data.unconverted_values,
+                    to_string_values=base_class_data.to_string_values
+                    | curr_base_class_data.to_string_values,
+                    parse_require=base_class_data.parse_require
+                    | curr_base_class_data.parse_require,
                 )
-                base_class_data.unconverted_values |= (
-                    curr_base_class_data.unconverted_values
-                )
-                base_class_data.to_string_values |= (
-                    curr_base_class_data.to_string_values
-                )
-                base_class_data.parse_require |= curr_base_class_data.parse_require
     return base_class_data
 
 
-def get_serial_class_data(type_class: type[Any]) -> SerialClassDataInspector:
+def get_serial_class_data(
+    type_class: type[ClassT],
+) -> SerialClassDataInspector[ClassT]:
     return SerialClassDataInspector(
-        _get_merged_serial_class_data(type_class) or EMPTY_SERIAL_CLASS_DATA
+        type_class, get_merged_serial_class_data(type_class) or EMPTY_SERIAL_CLASS_DATA
     )
 
 
@@ -138,13 +134,13 @@ class _SerialStringEnumData:
 
 
 def serial_string_enum(
-    *, labels: Optional[dict[str, str]] = None, deprecated: Optional[set[str]] = None
-) -> Callable[[_ClassT], _ClassT]:
+    *, labels: dict[str, str] | None = None, deprecated: set[str] | None = None
+) -> Callable[[ClassT], ClassT]:
     """
     A decorator for enums to provide serialization data, including labels.
     """
 
-    def decorate(orig_class: _ClassT) -> _ClassT:
+    def decorate(orig_class: ClassT) -> ClassT:
         cast(Any, orig_class).__unc_serial_string_enum_data = _SerialStringEnumData(
             labels=labels or {}, deprecated=deprecated or set()
         )
@@ -157,7 +153,7 @@ class SerialStringEnumInspector:
     def __init__(self, current: _SerialStringEnumData) -> None:
         self.current = current
 
-    def get_label(self, value: str) -> Optional[str]:
+    def get_label(self, value: str) -> str | None:
         return self.current.labels.get(value)
 
     def get_deprecated(self, value: str) -> bool:
