@@ -7,9 +7,11 @@ WORK-IN-PROGRESS, DON'T USE!
 import dataclasses
 import json
 import re
-from typing import Collection, cast
+from enum import StrEnum
+from typing import Collection, assert_never, cast
 
-import yaml
+from pkgs.serialization import yaml
+from pkgs.serialization_util import serialize_for_api
 
 from . import builder, util
 from .builder import EndpointGuideKey, RootGuideKey
@@ -61,6 +63,10 @@ base_name_map = {
 }
 
 
+class OpenAPIDefaultBehavior(StrEnum):
+    OPTIONAL_WITH_DEFAULT = "optional_with_default"
+
+
 def _rewrite_with_notice(
     file_path: str, file_content: str, *, notice: str = MODIFY_NOTICE
 ) -> bool:
@@ -125,7 +131,11 @@ def emit_open_api(builder: builder.SpecBuilder, *, config: OpenAPIConfig) -> Non
     for namespace in sorted(builder.namespaces.values(), key=lambda ns: ns.name):
         ctx = EmitOpenAPIContext(namespace=namespace)
 
-        if ctx.namespace.endpoint is not None and ctx.namespace.endpoint.is_beta:
+        if (
+            ctx.namespace.endpoint is not None
+            and ctx.namespace.endpoint.stability_level
+            == EmitOpenAPIStabilityLevel.draft
+        ):
             continue
 
         if ctx.namespace.name == "base":
@@ -169,7 +179,7 @@ def _serialize_global_context(ctx: EmitOpenAPIGlobalContext) -> str:
         oa_paths[path.path] = {"$ref": path.ref}
     oa_root["paths"] = oa_paths
 
-    return yaml.dump(oa_root, sort_keys=False)
+    return yaml.dumps(oa_root, sort_keys=False)
 
 
 def _is_empty_object_type(typ: OpenAPIType) -> bool:
@@ -251,18 +261,26 @@ def _emit_endpoint_parameters(
     } | _emit_endpoint_parameter_examples(examples)
 
 
-def _emit_is_beta(is_beta: bool) -> DictApiSchema:
-    if is_beta:
-        return {"x-beta": True}
-    return {}
-
-
 def _emit_stability_level(
     stability_level: EmitOpenAPIStabilityLevel | None,
 ) -> DictApiSchema:
-    if stability_level is not None:
-        return {"x-stability-level": str(stability_level)}
-    return {}
+    stability_info: dict[str, ApiSchema] = {}
+    resolved_stability_level = (
+        stability_level
+        if stability_level is not None
+        else EmitOpenAPIStabilityLevel.stable
+    )
+    stability_info["x-stability-level"] = str(resolved_stability_level)
+    match resolved_stability_level:
+        case EmitOpenAPIStabilityLevel.draft:
+            stability_info["x-beta"] = True
+        case EmitOpenAPIStabilityLevel.beta:
+            stability_info["x-badges"] = [{"name": "Beta", "color": "DarkOrange"}]
+        case EmitOpenAPIStabilityLevel.stable:
+            pass
+        case _:
+            assert_never(stability_level)
+    return stability_info
 
 
 def _emit_endpoint_request_body(
@@ -285,7 +303,9 @@ def _emit_endpoint_request_body(
                         "type": "object",
                         "title": "Body",
                         "required": ["data"],
-                        "properties": {"data": {"$ref": "#/components/schema/Arguments"}},
+                        "properties": {
+                            "data": {"$ref": "#/components/schema/Arguments"}
+                        },
                     }
                 }
                 | _emit_endpoint_argument_examples(examples)
@@ -357,10 +377,11 @@ def _emit_namespace(
                 "summary": endpoint.summary,
             }
             | _emit_endpoint_description(endpoint.description, ctx.endpoint.guides)
-            | _emit_is_beta(endpoint.is_beta)
             | _emit_stability_level(endpoint.stability_level)
             | _emit_endpoint_parameters(endpoint, argument_type, ctx.endpoint.examples)
-            | _emit_endpoint_request_body(endpoint, argument_type, ctx.endpoint.examples)
+            | _emit_endpoint_request_body(
+                endpoint, argument_type, ctx.endpoint.examples
+            )
             | {
                 "responses": {
                     "200": {
@@ -413,7 +434,19 @@ def _emit_namespace(
 
     path = f"{config.types_output}/common/{'/'.join(namespace.path)}.yaml"
     oa_namespace = {"components": oa_components}
-    _rewrite_with_notice(path, yaml.dump(oa_namespace, sort_keys=False))
+    _rewrite_with_notice(path, yaml.dumps(oa_namespace, sort_keys=False))
+
+
+def _get_openapi_default_behavior(
+    prop: builder.SpecProperty,
+) -> OpenAPIDefaultBehavior | None:
+    if prop.ext_info is None or prop.ext_info.get("open_api") is None:
+        return None
+    value_passed = prop.ext_info["open_api"].get("default_required_behavior")
+    if value_passed is None:
+        return None
+    assert isinstance(value_passed, str)
+    return OpenAPIDefaultBehavior(value_passed)
 
 
 def _emit_type(
@@ -479,6 +512,16 @@ def _emit_type(
                 # arguments, thus treat like extant==missing
                 # IMPROVE: if we can decide they are meant as output instead, then
                 # they should be marked as required
+                openapi_default_beahvior = _get_openapi_default_behavior(prop)
+                match openapi_default_beahvior:
+                    case None:
+                        pass
+                    case OpenAPIDefaultBehavior.OPTIONAL_WITH_DEFAULT:
+                        ref_type.nullable = True
+                        assert prop.default is not None, (
+                            "optional_with_default requires default"
+                        )
+                        ref_type.default = prop.default
                 properties[prop_name] = ref_type
             elif prop.extant == builder.PropertyExtant.missing:
                 # Unlike optional below, missing does not imply null is possible. They
@@ -558,7 +601,7 @@ def _emit_endpoint(
     ep = namespace.endpoint
     gctx.paths.append(
         EmitOpenAPIPath(
-            path=f"/{ep.path_root}/{ep.path_dirname}/{ep.path_basename}",
+            path=f"/{ep.resolved_path}",
             ref=ref_path,
         )
     )
@@ -568,20 +611,20 @@ def _emit_endpoint(
         description = f"**[External API-Endpoint]** <br/> {description}"
 
     path_cutoff = min(3, len(namespace.path) - 1)
+
     ctx.endpoint = EmitOpenAPIEndpoint(
         method=namespace.endpoint.method.lower(),
         tags=[tag_name],
         summary=f"{'/'.join(namespace.path[path_cutoff:])}",
         description=description,
-        is_beta=namespace.endpoint.is_beta,
         stability_level=namespace.endpoint.stability_level,
         examples=[
             EmitOpenAPIEndpointExample(
                 ref_name=f"ex_{i}",
                 summary=example.summary,
                 description=example.description,
-                arguments=example.arguments,
-                data=example.data,
+                arguments=serialize_for_api(example.arguments),
+                data=serialize_for_api(example.data),
             )
             for i, example in enumerate(endpoint_examples)
         ],
